TEXT_MAX_LENGTH = 2**32 - 1

BROWSERGYM_ID_ATTRIBUTE = "bid"  # Playwright's default is "data-testid"
BROWSERGYM_VISIBILITY_ATTRIBUTE = "browsergym_visibility_ratio"
BROWSERGYM_SETOFMARKS_ATTRIBUTE = "browsergym_set_of_marks"

EXTRACT_OBS_MAX_TRIES = 5
